// cargo klee --example registers --release

#![no_std]
#![no_main]

extern crate klee;
extern crate panic_abort;

extern crate volatile_register;
use volatile_register::RW;

// model RegA
// values always in the range 0..31
struct RegA(RW<u32>);

impl RegA {
  pub fn read(&self) -> u32 {
    let v = self.0.read();
    klee::kassume(v < 32);
    v
  }
  pub unsafe fn write(
    &self,
    v: u32,
  ) {
    klee::kassert!(v < 32);
    let v = self.0.write(v);
  }
}

// model RegB
// values always increases
struct RegB(RW<u32>);

static mut STATE: u32 = 0; // reset value
impl RegB {
  pub fn read(&self) -> u32 {
    unsafe {
      STATE += 1;
      STATE
    }
  }
  pub unsafe fn write(
    &self,
    v: u32,
  ) {
    STATE = v;
  }
}

#[no_mangle]
fn main() {
  let reg_a: RegA =
    unsafe { core::mem::uninitialized() };
  let reg_b: RegB =
    unsafe { core::mem::uninitialized() };

  unsafe {
    reg_b.write(0);
  }

  let t1 = reg_b.read();
  let t2 = reg_b.read();

  let diff = 100 / (t2 - t1);

  klee::kassert!(t2 > t1);

  let v = reg_a.read();
  if v > 5 {
    unsafe {
      reg_a.write(v - 1);
    }
    if reg_a.read() == v - 1 {
      panic!();
    }
  }
}
