#![allow(non_snake_case)]
#![no_main]

use klee::{kassert, ksymbol};

#[no_mangle]
fn main() {
  let mut DATA: [Data; (DISCREPENCY + 1) as usize] =
    unsafe { core::mem::uninitialized() };
  ksymbol!(&mut DATA, "DATA");

  let mut TIMEOUT_CNTR: u32 =
    unsafe { core::mem::uninitialized() };
  ksymbol!(&mut TIMEOUT_CNTR, "TIMEOUT_CNTR");

  let mut STATE_HISTORY: [State;
    (DISCREPENCY + 1) as usize] =
    unsafe { core::mem::uninitialized() };

  DATA[0] = Data { a: false, b: false };
  TIMEOUT_CNTR = 0;
  let mut STATE = S8001;

  for i in 0..(DISCREPENCY + 1) as usize {
    periodic(
      &mut STATE,
      &mut TIMEOUT_CNTR,
      &mut DATA[i as usize],
    );

    // invariants
    // S8000 -> a & b
    if STATE == S8000 {
      kassert!(DATA[i].a & DATA[i].b);
    }
    // !a | !b -> ! S8000
    if !DATA[i].a | !DATA[i].b {
      kassert!(!(STATE == S8000));
    }
    // S8001 -> !a & !b
    if STATE == S8001 {
      kassert!(!DATA[i].a | !DATA[i].b);
    }
    // !a & !b -> S8001
    if !DATA[i].a & !DATA[i].b {
      kassert!(STATE == S8001);
    }

    // error states implies timeout
    match STATE {
      C001 | C002 | C002 => kassert!(TIMEOUT_CNTR == 0),
      _ => {}
    }

    if i > 0 {
      match STATE_HISTORY[i - 1] {
        C001 | C002 | C003 | S8005 => {
          kassert!(!(STATE == S8000));
        }
        _ => {
          if DATA[i].a & DATA[i].b {
            kassert!(STATE == S8000);
          }
        }
      }
    }
    STATE_HISTORY[i] = STATE;
  }
}

#[derive(Debug, Copy, Clone)]
pub struct Data {
  a: bool,
  b: bool,
}

#[derive(Debug, Copy, Clone, PartialEq)]
pub enum State {
  S8000,
  S8001,
  S8004,
  S8014,
  S8005,
  C001,
  C002,
  C003,
}

use State::*;

const DISCREPENCY: u32 = 4;
const T: bool = true;
const F: bool = false;

fn periodic(
  STATE: &mut State,
  TIMEOUT_CNTR: &mut u32,
  DATA: &mut Data,
) {
  // we start directly in the init state S80001
  // static mut STATE: State = State::S8001;
  // static mut TIMEOUT_CNTR: u32 = 0;

  let data = DATA;

  *STATE = match STATE {
    S8000 => match (data.a, data.b) {
      (F, F) => S8001,
      (F, T) | (T, F) => {
        *TIMEOUT_CNTR = DISCREPENCY;
        S8005
      }
      (T, T) => S8000,
    },
    S8001 => match (data.a, data.b) {
      (F, F) => S8001,
      (F, T) => {
        *TIMEOUT_CNTR = DISCREPENCY;
        S8014
      }
      (T, F) => {
        *TIMEOUT_CNTR = DISCREPENCY;
        S8004
      }
      (T, T) => S8000,
    },
    S8004 => {
      *TIMEOUT_CNTR -= 1;
      match *TIMEOUT_CNTR {
        0 => C001,
        _ => match (data.a, data.b) {
          (F, _) => S8001,
          (_, T) => S8000,
          _ => S8004,
        },
      }
    }
    S8014 => {
      *TIMEOUT_CNTR -= 1;
      match *TIMEOUT_CNTR {
        0 => C002,
        _ => match (data.a, data.b) {
          (_, F) => S8001,
          (T, _) => S8000,
          _ => S8014,
        },
      }
    }
    S8005 => {
      *TIMEOUT_CNTR -= 1;
      match *TIMEOUT_CNTR {
        0 => C003,
        _ => match (data.a, data.b) {
          (F, F) => S8001,
          _ => S8005,
        },
      }
    }
    C001 | C002 => match (data.a, data.b) {
      (F, F) => S8001,
      _ => *STATE,
    },
    C003 => match (data.a, data.b) {
      (F, F) => S8001,
      _ => C003,
    },
  };
}
